try:
    import appium
    appium_installed = True
except ImportError:
    appium_installed = False

from .webdriver import WebDriver, WebDriverType, Browser, BrowserType
from .staticelement import StaticElement
from .identifier import Identifier
from .waiter import Waiter
from .exceptions import EasyiumException, TimeoutException, NoSuchElementException, NotPersistException, \
    LatePersistException, UnsupportedWebDriverTypeException, InvalidByException, UnsupportedOperationForWebDriver
from .config import default_config

__author__ = 'karl.gong'



from .element import Element

__author__ = 'karl.gong'


class StaticElement(Element):
    def __init__(self, parent, locator):
        """
            Creates a new instance of the StaticElement.

        :param parent: the parent context
        :param locator: the locator of this element (relative to parent context)
        """
        Element.__init__(self, parent)
        self.__selenium_element = None
        self.__locator = locator

    def _selenium_context(self):
        if self.__selenium_element is None:
            self._refresh()
        return self.__selenium_element

    def _selenium_element(self):
        if self.__selenium_element is None:
            self._refresh()
        return self.__selenium_element

    def _refresh(self):
        self.__selenium_element = self.get_parent()._find_selenium_element(self.__locator)

    def persist(self):
        self.get_parent().persist()

    def __str__(self):
        if self.__selenium_element is None:
            element_id = None
        else:
            element_id = self.__selenium_element.id
        return "%s\n|- StaticElement [SeleniumElement: %s][Locator: %s]" % (self.get_parent(), element_id, self.__locator)



__author__ = 'karl.gong'


class EasyiumException(Exception):
    pass


class TimeoutException(EasyiumException):
    pass


class NoSuchElementException(EasyiumException):
    pass


class NotPersistException(EasyiumException):
    pass


class LatePersistException(EasyiumException):
    pass


class UnsupportedWebDriverTypeException(EasyiumException):
    pass


class InvalidByException(EasyiumException):
    pass


class UnsupportedOperationForWebDriver(EasyiumException):
    pass


from selenium.webdriver import Ie, Firefox, Chrome, Opera, Safari, PhantomJS
from selenium.common.exceptions import NoAlertPresentException

from .context import Context
from .waiter import WebDriverWaitFor
from .exceptions import UnsupportedWebDriverTypeException
from .config import DEFAULT, default_config
from .decorator import SupportedBy

__author__ = 'karl.gong'


class WebDriverType:
    IE = "ie"
    FIREFOX = "firefox"
    CHROME = "chrome"
    OPERA = "opera"
    SAFARI = "safari"
    PHANTOMJS = "phantomjs"
    ANDROID = "android"
    IOS = "ios"

    _BROWSER = [IE, FIREFOX, CHROME, OPERA, SAFARI, PHANTOMJS]
    _MOBILE = [ANDROID, IOS]
    _ALL = _BROWSER + _MOBILE


class WebDriver(Context):
    def __init__(self, web_driver_type=WebDriverType.CHROME, page_load_timeout=DEFAULT, script_timeout=DEFAULT,
                 wait_interval=DEFAULT, wait_timeout=DEFAULT, **kwargs):
        """
            Creates a new instance of the WebDriver.

        :param web_driver_type: the web driver type
        :param page_load_timeout: the page load timeout (in milliseconds), default value is from default_config.web_driver_page_load_timeout
        :param script_timeout: the script timeout (in milliseconds), default value is from default_config.web_driver_script_timeout
        :param wait_interval: the wait interval (in milliseconds), default value is from default_config.web_driver_wait_interval
        :param wait_timeout: the wait timeout (in milliseconds), default value is from default_config.web_driver_wait_timeout
        :param kwargs: the keyword args for the web driver specified by web_driver_type
        """
        Context.__init__(self)
        self.__web_driver_type = web_driver_type.lower()
        if self.__web_driver_type == WebDriverType.IE:
            self.__selenium_web_driver = Ie(**kwargs)
        elif self.__web_driver_type == WebDriverType.FIREFOX:
            self.__selenium_web_driver = Firefox(**kwargs)
        elif self.__web_driver_type == WebDriverType.CHROME:
            self.__selenium_web_driver = Chrome(**kwargs)
        elif self.__web_driver_type == WebDriverType.OPERA:
            self.__selenium_web_driver = Opera(**kwargs)
        elif self.__web_driver_type == WebDriverType.SAFARI:
            self.__selenium_web_driver = Safari(**kwargs)
        elif self.__web_driver_type == WebDriverType.PHANTOMJS:
            self.__selenium_web_driver = PhantomJS(**kwargs)
        elif self.__web_driver_type in WebDriverType._MOBILE:
            from appium.webdriver.webdriver import WebDriver as Mobile
            self.__selenium_web_driver = Mobile(**kwargs)
        else:
            raise UnsupportedWebDriverTypeException("The web driver type [%s] is not supported." % web_driver_type)
        self.set_page_load_timeout(default_config.web_driver_page_load_timeout if page_load_timeout == DEFAULT else page_load_timeout)
        self.set_script_timeout(default_config.web_driver_script_timeout if script_timeout == DEFAULT else script_timeout)
        self.__wait_interval = default_config.web_driver_wait_interval if wait_interval == DEFAULT else wait_interval
        self.__wait_timeout = default_config.web_driver_wait_timeout if wait_timeout == DEFAULT else wait_timeout

    def _selenium_web_driver(self):
        return self.__selenium_web_driver

    def _selenium_context(self):
        return self.__selenium_web_driver

    def get_web_driver(self):
        return self

    get_browser = get_web_driver

    def get_web_driver_type(self):
        return self.__web_driver_type

    get_browser_type = get_web_driver_type

    def get_wait_interval(self):
        return self.__wait_interval

    def get_wait_timeout(self):
        return self.__wait_timeout

    def wait_for(self, interval=DEFAULT, timeout=DEFAULT):
        """
            Get a WebDriverWaitFor instance.

        :param interval: the wait interval (in milliseconds), default value is web driver's wait interval
        :param timeout: the wait timeout (in milliseconds), default value is web driver's wait timeout
        """
        interval = self.get_wait_interval() if interval == DEFAULT else interval
        timeout = self.get_wait_timeout() if timeout == DEFAULT else timeout
        return WebDriverWaitFor(self, interval, timeout)

    @SupportedBy(WebDriverType._BROWSER)
    def maximize_window(self):
        self.__selenium_web_driver.maximize_window()

    def set_page_load_timeout(self, timeout):
        self.__selenium_web_driver.set_page_load_timeout(timeout / 1000.0)

    def set_script_timeout(self, timeout):
        self.__selenium_web_driver.set_script_timeout(timeout / 1000.0)

    def execute_script(self, script, *args):
        from .element import Element

        converted_args = []
        for arg in args:
            if isinstance(arg, Element):
                converted_args.append(arg._selenium_element())
            else:
                converted_args.append(arg)

        return self.__selenium_web_driver.execute_script(script, *converted_args)

    def execute_async_script(self, script, *args):
        from .element import Element

        converted_args = []
        for arg in args:
            if isinstance(arg, Element):
                converted_args.append(arg._selenium_element())
            else:
                converted_args.append(arg)

        return self.__selenium_web_driver.execute_async_script(script, *converted_args)

    def open(self, url):
        self.__selenium_web_driver.get(url)

    get = open

    def get_title(self):
        return self.__selenium_web_driver.title

    def refresh(self):
        self.__selenium_web_driver.refresh()

    def back(self):
        self.__selenium_web_driver.back()

    def forward(self):
        self.__selenium_web_driver.forward()

    def quit(self):
        self.__selenium_web_driver.quit()

    def get_current_url(self):
        return self.__selenium_web_driver.current_url

    def get_page_source(self):
        return self.__selenium_web_driver.page_source

    @SupportedBy(WebDriverType._MOBILE)
    def get_contexts(self):
        return self.__selenium_web_driver.contexts

    @SupportedBy(WebDriverType._MOBILE)
    def get_current_context(self):
        return self.__selenium_web_driver.current_context

    @SupportedBy(WebDriverType._MOBILE)
    def switch_to_context(self, context_name):
        self.__selenium_web_driver.switch_to.context(context_name)

    def switch_to_frame(self, frame_reference):
        self.__selenium_web_driver.switch_to.frame(frame_reference)

    def switch_to_parent_frame(self):
        self.__selenium_web_driver.switch_to.parent_frame()

    def switch_to_default_content(self):
        self.__selenium_web_driver.switch_to.default_content()

    def get_alert(self):
        return self.__selenium_web_driver.switch_to.alert

    def is_alert_present(self):
        try:
            alert_text = self.__selenium_web_driver.switch_to.alert.text
            return True
        except NoAlertPresentException:
            return False

    def get_cookies(self):
        return self.__selenium_web_driver.get_cookies()

    def get_cookie(self, name):
        return self.__selenium_web_driver.get_cookie(name)

    def delete_cookie(self, name):
        self.__selenium_web_driver.delete_cookie(name)

    def delete_all_cookies(self):
        self.__selenium_web_driver.delete_all_cookies()

    def add_cookie(self, cookie_dict):
        self.__selenium_web_driver.add_cookie(cookie_dict)

    def get_desired_capabilities(self):
        return self.__selenium_web_driver.desired_capabilities

    def get_screenshot_as_file(self, filename):
        return self.__selenium_web_driver.get_screenshot_as_file(filename)

    def get_screenshot_as_png(self):
        return self.__selenium_web_driver.get_screenshot_as_png()

    def get_screenshot_as_base64(self):
        return self.__selenium_web_driver.get_screenshot_as_base64()

    save_screenshot = get_screenshot_as_file

    def get_current_window_handle(self):
        return self.__selenium_web_driver.current_window_handle

    def get_window_handles(self):
        return self.__selenium_web_driver.window_handles

    def switch_to_window(self, window_reference):
        self.__selenium_web_driver.switch_to.window(window_reference)

    def close_window(self, window_reference="current"):
        if window_reference == "current":
            self.__selenium_web_driver.close()
        else:
            current_window = self.get_current_window_handle()
            self.switch_to_window(window_reference)
            self.__selenium_web_driver.close()
            self.switch_to_window(current_window)

    def set_window_size(self, width, height, window_reference="current"):
        self.__selenium_web_driver.set_window_size(width, height, window_reference)

    def get_window_size(self, window_reference="current"):
        return self.__selenium_web_driver.get_window_size(window_reference)

    def set_window_position(self, x, y, window_reference="current"):
        self.__selenium_web_driver.set_window_position(x, y, window_reference)

    def get_window_position(self, window_reference="current"):
        return self.__selenium_web_driver.get_window_position(window_reference)

    def get_orientation(self):
        return self.__selenium_web_driver.orientation

    def set_orientation(self, value):
        self.__selenium_web_driver.orientation = value.upper()

    def get_application_cache(self):
        return self.__selenium_web_driver.application_cache

    def get_log_types(self):
        return self.__selenium_web_driver.log_types

    def get_log(self, log_type):
        return self.__selenium_web_driver.get_log(log_type)

    @SupportedBy(WebDriverType._MOBILE)
    def swipe(self, start_x, start_y, end_x, end_y, duration=None):
        self.__selenium_web_driver.swipe(start_x, start_y, end_x, end_y, duration)

    @SupportedBy(WebDriverType._MOBILE)
    def flick(self, start_x, start_y, end_x, end_y):
        self.__selenium_web_driver.flick(start_x, start_y, end_x, end_y)

    @SupportedBy(WebDriverType._MOBILE)
    def app_strings(self, language=None, string_file=None):
        return self.__selenium_web_driver.app_strings(language, string_file)

    @SupportedBy(WebDriverType._MOBILE)
    def reset_current_app(self):
        self.__selenium_web_driver.reset()

    @SupportedBy(WebDriverType._MOBILE)
    def hide_keyboard(self, key_name=None, key=None, strategy=None):
        self.__selenium_web_driver.hide_keyboard(key_name, key, strategy)

    @SupportedBy(WebDriverType.ANDROID)
    def key_event(self, key_code, meta_state=None):
        self.__selenium_web_driver.keyevent(key_code, meta_state)

    @SupportedBy(WebDriverType.ANDROID)
    def press_key(self, key_code, meta_state=None):
        self.__selenium_web_driver.press_keycode(key_code, meta_state)

    @SupportedBy(WebDriverType.ANDROID)
    def long_press_key(self, key_code, meta_state=None):
        self.__selenium_web_driver.long_press_keycode(key_code, meta_state)

    @SupportedBy(WebDriverType.ANDROID)
    def get_current_activity(self):
        return self.__selenium_web_driver.current_activity

    @SupportedBy(WebDriverType._MOBILE)
    def pull_file(self, path):
        return self.__selenium_web_driver.pull_file(path)

    @SupportedBy(WebDriverType._MOBILE)
    def pull_folder(self, path):
        return self.__selenium_web_driver.pull_folder(path)

    @SupportedBy(WebDriverType._MOBILE)
    def push_file(self, path, base64data):
        self.__selenium_web_driver.push_file(path, base64data)

    @SupportedBy(WebDriverType._MOBILE)
    def send_to_background(self, duration):
        self.__selenium_web_driver.background_app(duration / 1000.0)

    @SupportedBy(WebDriverType._MOBILE)
    def is_app_installed(self, bundle_id):
        return self.__selenium_web_driver.is_app_installed(bundle_id)

    @SupportedBy(WebDriverType._MOBILE)
    def install_app(self, app_path):
        self.__selenium_web_driver.install_app(app_path)

    @SupportedBy(WebDriverType._MOBILE)
    def remove_app(self, app_id):
        self.__selenium_web_driver.remove_app(app_id)

    @SupportedBy(WebDriverType._MOBILE)
    def launch_app(self):
        self.__selenium_web_driver.launch_app()

    @SupportedBy(WebDriverType._MOBILE)
    def close_app(self):
        self.__selenium_web_driver.close_app()

    @SupportedBy(WebDriverType.ANDROID)
    def start_activity(self, app_package, app_activity, app_wait_package=DEFAULT, app_wait_activity=DEFAULT,
                       intent_action=DEFAULT, intent_category=DEFAULT, intent_flags=DEFAULT,
                       optional_intent_arguments=DEFAULT, stop_app_on_reset=DEFAULT):
        options = {}
        if app_wait_package != DEFAULT:
            options["app_wait_package"] = app_wait_package
        if app_wait_activity != DEFAULT:
            options["app_wait_activity"] = app_wait_activity
        if intent_action != DEFAULT:
            options["intent_action"] = intent_action
        if intent_category != DEFAULT:
            options["intent_category"] = intent_category
        if intent_flags != DEFAULT:
            options["intent_flags"] = intent_flags
        if optional_intent_arguments != DEFAULT:
            options["optional_intent_arguments"] = optional_intent_arguments
        if stop_app_on_reset != DEFAULT:
            options["stop_app_on_reset"] = stop_app_on_reset

        self.__selenium_web_driver.start_activity(app_package, app_activity, **options)

    @SupportedBy(WebDriverType.ANDROID)
    def end_test_coverage(self, intent, path):
        self.__selenium_web_driver.end_test_coverage(intent, path)

    @SupportedBy(WebDriverType.IOS)
    def lock(self, duration):
        self.__selenium_web_driver.lock(duration / 1000.0)

    @SupportedBy(WebDriverType._MOBILE)
    def shake(self):
        self.__selenium_web_driver.shake()

    @SupportedBy(WebDriverType.ANDROID)
    def open_notifications(self):
        self.__selenium_web_driver.open_notifications()

    @SupportedBy(WebDriverType.ANDROID)
    def get_network_connection(self):
        return self.__selenium_web_driver.network_connection

    @SupportedBy(WebDriverType.ANDROID)
    def set_network_connection(self, connection_type):
        self.__selenium_web_driver.set_network_connection(connection_type)

    @SupportedBy(WebDriverType.ANDROID)
    def get_available_ime_engines(self):
        return self.__selenium_web_driver.available_ime_engines

    @SupportedBy(WebDriverType.ANDROID)
    def is_ime_service_active(self):
        return self.__selenium_web_driver.is_ime_active()

    @SupportedBy(WebDriverType.ANDROID)
    def active_ime_engine(self, engine):
        self.__selenium_web_driver.activate_ime_engine(engine)

    @SupportedBy(WebDriverType.ANDROID)
    def deactivate_current_ime_engine(self):
        self.__selenium_web_driver.deactivate_ime_engine()

    @SupportedBy(WebDriverType.ANDROID)
    def get_current_ime_engine(self):
        return self.__selenium_web_driver.active_ime_engine

    @SupportedBy(WebDriverType._MOBILE)
    def get_settings(self):
        return self.__selenium_web_driver.get_settings()

    @SupportedBy(WebDriverType._MOBILE)
    def update_settings(self, settings):
        self.__selenium_web_driver.update_settings(settings)

    @SupportedBy(WebDriverType.ANDROID)
    def toggle_location_services(self):
        self.__selenium_web_driver.toggle_location_services()

    @SupportedBy(WebDriverType._MOBILE)
    def set_location(self, latitude, longitude, altitude):
        self.__selenium_web_driver.set_location(latitude, longitude, altitude)

    def __str__(self):
        return "WebDriver [WebDriverType: %s][SessionId: %s]" % (self.__selenium_web_driver.name, self.__selenium_web_driver.session_id)


Browser = WebDriver
BrowserType = WebDriverType



__author__ = 'karl.gong'


class Identifier:
    @staticmethod
    def id(element):
        return "id=" + element.get_attribute("id")

    @staticmethod
    def class_name(element):
        return "class=" + element.get_attribute("class")

    @staticmethod
    def name(element):
        return "name=" + element.get_attribute("name")

    @staticmethod
    def link_text(element):
        return "link=" + element.get_text()



from selenium.common.exceptions import StaleElementReferenceException, NoSuchElementException, WebDriverException

from .locator import locator_to_by_value
from .identifier import Identifier
from .waiter import Waiter
from .config import DEFAULT
from . import exceptions

__author__ = 'karl.gong'


class Context:
    def __init__(self):
        pass

    def get_web_driver(self):
        pass

    def get_web_driver_type(self):
        pass

    def get_wait_interval(self):
        pass

    def get_wait_timeout(self):
        pass

    def waiter(self, interval=DEFAULT, timeout=DEFAULT):
        """"
            Get a Waiter instance.

        :param interval: the wait interval (in milliseconds), default value is web driver's wait interval
        :param timeout: the wait timeout (in milliseconds), default value is web driver's wait timeout
        """
        interval = self.get_wait_interval() if interval == DEFAULT else interval
        timeout = self.get_wait_timeout() if timeout == DEFAULT else timeout
        return Waiter(interval, timeout)

    def _selenium_context(self):
        pass

    def _refresh(self):
        pass

    def persist(self):
        pass

    def _find_selenium_element(self, locator):
        by, value = locator_to_by_value(locator)
        try:
            try:
                return self._selenium_context().find_element(by, value)
            except StaleElementReferenceException:
                self._refresh()
                return self._selenium_context().find_element(by, value)
        except NoSuchElementException:
            raise exceptions.NoSuchElementException("Cannot find element by [%s] under:\n%s\n" % (locator, self))
        except WebDriverException as wde:
            raise exceptions.EasyiumException("%s\n%s" % (wde.msg, self))

    def find_element(self, locator, identifier=Identifier.id):
        """
            Find a DynamicElement under this context immediately.
            If this context is not existing, it will raise NoSuchElementException.

        :param locator: the locator to find the DynamicElement (relative to parent context)
        :param identifier: the identifier to refresh the DynamicElement
        :return the DynamicElement found by locator
        """
        # import the DynamicElement here to avoid cyclic dependency
        from .dynamicelement import DynamicElement

        by, value = locator_to_by_value(locator)
        try:
            try:
                return DynamicElement(self, self._selenium_context().find_element(by, value), locator, identifier)
            except StaleElementReferenceException:
                self._refresh()
                return DynamicElement(self, self._selenium_context().find_element(by, value), locator, identifier)
        except NoSuchElementException:
            raise exceptions.NoSuchElementException("Cannot find element by [%s] under:\n%s\n" % (locator, self))
        except WebDriverException as wde:
            raise exceptions.EasyiumException("%s\n%s" % (wde.msg, self))

    def find_elements(self, locator, identifier=Identifier.id):
        """
            Find DynamicElement list under this context immediately.
            If this context is not existing, it will raise NoSuchElementException.

        :param locator: the locator to find the DynamicElement list (relative to parent context)
        :param identifier: the identifier to refresh the DynamicElement
        :return the DynamicElement list found by locator
        """
        # import the DynamicElement here to avoid cyclic dependency
        from .dynamicelement import DynamicElement

        by, value = locator_to_by_value(locator)
        try:
            try:
                selenium_elements = self._selenium_context().find_elements(by, value)
            except StaleElementReferenceException:
                self._refresh()
                selenium_elements = self._selenium_context().find_elements(by, value)
        except WebDriverException as wde:
            raise exceptions.EasyiumException("%s\n%s" % (wde.msg, self))

        elements = []
        for selenium_element in selenium_elements:
            elements.append(DynamicElement(self, selenium_element, locator, identifier))
        return elements



__author__ = 'karl.gong'

# place holder for using default value
DEFAULT = "DEFAULT"


class Config:
    def __init__(self):
        # default configuration for web driver
        self.web_driver_page_load_timeout = 30000
        self.web_driver_script_timeout = 30000
        self.web_driver_wait_interval = 1000
        self.web_driver_wait_timeout = 30000

        # default configuration for waiter
        self.waiter_wait_interval = 1000
        self.waiter_wait_timeout = 30000


default_config = Config()



__author__ = 'karl.gong'

from .exceptions import UnsupportedOperationForWebDriver


def SupportedBy(*web_driver_types):
    def handle_func(func):
        def handle_args(*args, **kwargs):
            wd_types = []
            for wd_type in web_driver_types:
                if isinstance(wd_type, list):
                    wd_types += wd_type
                else:
                    wd_types += [wd_type]

            current_web_driver_type = args[0].get_web_driver().get_web_driver_type()
            if current_web_driver_type not in wd_types:
                raise UnsupportedOperationForWebDriver(
                    "The operation is not supported in web driver [%s]." % current_web_driver_type)

            return func(*args, **kwargs)

        return handle_args

    return handle_func



from .exceptions import NotPersistException, LatePersistException
from .element import Element

__author__ = 'karl.gong'


class DynamicElement(Element):
    def __init__(self, parent, selenium_element, found_by, identifier):
        Element.__init__(self, parent)
        self.__selenium_element = selenium_element
        self.__locator = None
        self.__found_by = found_by
        self.__identifier = identifier

    def _selenium_context(self):
        return self.__selenium_element

    def _selenium_element(self):
        return self.__selenium_element

    def _refresh(self):
        if self.__locator is None:
            raise NotPersistException("persist() was not invoked so this Element cannot auto-refresh.\n%s" % self)
        self.__selenium_element = self.get_parent()._find_selenium_element(self.__locator)

    def persist(self):
        self.get_parent().persist()

        try:
            if self.__locator is None:
                self.__locator = self.__identifier(self)
        except NotPersistException:
            raise LatePersistException(
                "Trying to persist() a stale element. Try invoking persist() earlier.\n%s" % self)

    def __str__(self):
        return "%s\n|- DynamicElement [SeleniumElement: %s][Locator: %s][FoundBy: %s]" % (
            self.get_parent(), self.__selenium_element.id, self.__locator, self.__found_by)



from selenium.webdriver.common.by import By

from .exceptions import InvalidByException
from . import appium_installed

__author__ = 'karl.gong'

locator_to_by_map = {
    "id": By.ID,
    "xpath": By.XPATH,
    "link": By.LINK_TEXT,
    "partial_link": By.PARTIAL_LINK_TEXT,
    "name": By.NAME,
    "tag": By.TAG_NAME,
    "class": By.CLASS_NAME,
    "css": By.CSS_SELECTOR
}

if appium_installed:
    from appium.webdriver.common.mobileby import MobileBy

    mobile_locator_to_by_map = {
        "ios_uiautomation": MobileBy.IOS_UIAUTOMATION,
        "android_uiautomation": MobileBy.ANDROID_UIAUTOMATOR,
        "accessibility_id": MobileBy.ACCESSIBILITY_ID
    }

    locator_to_by_map.update(mobile_locator_to_by_map)


def locator_to_by_value(locator):
    separator_index = locator.find("=")
    by = locator[:separator_index]
    value = locator[separator_index + 1:]
    try:
        by = locator_to_by_map[by]
    except KeyError:
        raise InvalidByException("Invalid By: [%s] for locator: [%s]." % (by, locator))
    return by, value


from selenium.common.exceptions import WebDriverException, StaleElementReferenceException
from selenium.webdriver import ActionChains

from .decorator import SupportedBy
from .webdriver import WebDriverType
from .exceptions import EasyiumException, NoSuchElementException
from .waiter import ElementWaitFor
from .context import Context
from .config import DEFAULT

__author__ = 'karl.gong'


class Element(Context):
    def __init__(self, parent):
        Context.__init__(self)
        self.__parent = parent

    def get_web_driver(self):
        return self.get_parent().get_web_driver()

    get_browser = get_web_driver

    def get_web_driver_type(self):
        return self.get_web_driver().get_web_driver_type()

    get_browser_type = get_web_driver_type

    def get_wait_interval(self):
        return self.get_web_driver().get_wait_interval()

    def get_wait_timeout(self):
        return self.get_web_driver().get_wait_timeout()

    def get_parent(self):
        return self.__parent

    def _selenium_element(self):
        pass

    def wait_for(self, interval=DEFAULT, timeout=DEFAULT):
        """
            Get a ElementWaitFor instance.

        :param interval: the wait interval (in milliseconds), default value is web driver's wait interval
        :param timeout: the wait timeout (in milliseconds), default value is web driver's wait timeout
        """
        interval = self.get_wait_interval() if interval == DEFAULT else interval
        timeout = self.get_wait_timeout() if timeout == DEFAULT else timeout
        return ElementWaitFor(self, interval, timeout)

    def blur(self):
        try:
            try:
                self.get_web_driver().execute_script("arguments[0].blur()", self)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                self.get_web_driver().execute_script("arguments[0].blur()", self)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def clear(self):
        try:
            try:
                self._selenium_element().clear()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                self._selenium_element().clear()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def click(self):
        try:
            try:
                self._selenium_element().click()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                self._selenium_element().click()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def send_keys(self, value):
        try:
            try:
                self._selenium_element().send_keys(value)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                self._selenium_element().send_keys(value)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def submit(self):
        try:
            try:
                self._selenium_element().submit()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                self._selenium_element().submit()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_attribute(self, name):
        try:
            try:
                return self._selenium_element().get_attribute(name)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                return self._selenium_element().get_attribute(name)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_css_value(self, property_name):
        try:
            try:
                return self._selenium_element().value_of_css_property(property_name)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                return self._selenium_element().value_of_css_property(property_name)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_location(self):
        """
            Usage: x, y = element.get_location()
        :return: The location for the top-left corner of this element.
        """
        try:
            try:
                location = self._selenium_element().location
                return location["x"], location["y"]
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                location = self._selenium_element().location
                return location["x"], location["y"]
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_size(self):
        """
            Usage: width, height = element.get_size()
        :return: The size (including border) of this element.
        """
        try:
            try:
                size = self._selenium_element().size
                return size["width"], size["height"]
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                size = self._selenium_element().size
                return size["width"], size["height"]
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_tag_name(self):
        try:
            try:
                return self._selenium_element().tag_name
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                return self._selenium_element().tag_name
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_value(self):
        try:
            try:
                return self._selenium_element().get_attribute("value")
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                return self._selenium_element().get_attribute("value")
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_text(self):
        try:
            try:
                return self._selenium_element().text
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                return self._selenium_element().text
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_text_node_content(self, text_node_index):
        """
            Get content of the text node in this element.
            If the text_node_index refers to a non-text node or be out of bounds, an exception will be thrown.
        :param text_node_index: index of text node in this element
        :return: the content of the text node in this element.
        """
        try:
            try:
                content = self.get_web_driver().execute_script(
                    "return arguments[0].childNodes[%s].nodeValue" % text_node_index, self)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                content = self.get_web_driver().execute_script(
                    "return arguments[0].childNodes[%s].nodeValue" % text_node_index, self)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

        if content is None:
            raise EasyiumException("Cannot get text content of a non-text node in element: \n%s\n" % self)
        return content

    def set_selection_range(self, start, end):
        """
            Set the selection range for text in this element..
        :param start: start position
        :param end: end position
        """
        script = """
            function getTextNodesIn(node) {
                var textNodes = [];
                if (node.nodeType == 3) {
                    textNodes.push(node);
                } else {
                    var children = node.childNodes;
                    for (var i = 0, len = children.length; i < len; ++i) {
                        textNodes.push.apply(textNodes, getTextNodesIn(children[i]));
                    }
                }
                return textNodes;
            }

            function setSelectionRange(el, start, end) {
                if (el.tagName == 'INPUT' || el.tagName == 'TEXTAREA'){
                    if(el.createTextRange){
                        var Range=el.createTextRange();
                        Range.collapse();
                        Range.moveEnd('character',end);
                        Range.moveStart('character',start);
                        Range.select();
                    }else if(el.setSelectionRange){
                        el.focus();
                        el.setSelectionRange(start,end);
                    }
                } else {
					if (document.createRange && window.getSelection) {
                        var range = document.createRange();
                        range.selectNodeContents(el);
                        var textNodes = getTextNodesIn(el);
                        var foundStart = false;
                        var charCount = 0, endCharCount;

                        for (var i = 0, textNode; textNode = textNodes[i++]; ) {
                            endCharCount = charCount + textNode.length;
                            if (!foundStart && start >= charCount
                                    && (start < endCharCount ||
                                    (start == endCharCount && i < textNodes.length))) {
                                range.setStart(textNode, start - charCount);
                                foundStart = true;
                            }
                            if (foundStart && end <= endCharCount) {
                                range.setEnd(textNode, end - charCount);
                                break;
                            }
                            charCount = endCharCount;
                        }

                        var sel = window.getSelection();
                        sel.removeAllRanges();
                        sel.addRange(range);
                    } else if (document.selection && document.body.createTextRange) {
                        var textRange = document.body.createTextRange();
                        textRange.moveToElementText(el);
                        textRange.collapse(true);
                        textRange.moveEnd('character', end);
                        textRange.moveStart('character', start);
                        textRange.select();
                    }
				}
            }

            setSelectionRange(arguments[0], %s, %s);
        """
        try:
            try:
                self.get_web_driver().execute_script(script % (start, end), self)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                self.get_web_driver().execute_script(script % (start, end), self)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def get_inner_html(self):
        try:
            try:
                return self.get_web_driver().execute_script("return arguments[0].innerHTML", self)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                return self.get_web_driver().execute_script("return arguments[0].innerHTML", self)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def is_enabled(self):
        try:
            try:
                return self._selenium_element().is_enabled()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                return self._selenium_element().is_enabled()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def is_selected(self):
        try:
            try:
                return self._selenium_element().is_selected()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                return self._selenium_element().is_selected()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._BROWSER)
    def mouse_over(self):
        script = """
            var evObj = document.createEvent('MouseEvents');
            evObj.initMouseEvent("mouseover",true, false, window, 0, 0, 0, 0, 0, false, false, false, false, 0, null);
            arguments[0].dispatchEvent(evObj);
        """
        web_driver_type = self.get_web_driver_type()
        try:
            try:
                if web_driver_type == WebDriverType.CHROME or web_driver_type == WebDriverType.FIREFOX:
                    self.get_web_driver().execute_script(script, self)
                else:
                    ActionChains(self.get_web_driver()._selenium_web_driver()).move_to_element(
                        self._selenium_element()).perform()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                if web_driver_type == WebDriverType.CHROME or web_driver_type == WebDriverType.FIREFOX:
                    self.get_web_driver().execute_script(script, self)
                else:
                    ActionChains(self.get_web_driver()._selenium_web_driver()).move_to_element(
                        self._selenium_element()).perform()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._BROWSER)
    def mouse_out(self):
        script = """
            var evObj = document.createEvent('MouseEvents');
            evObj.initMouseEvent("mouseout",true, false, window, 0, 0, 0, 0, 0, false, false, false, false, 0, null);
            arguments[0].dispatchEvent(evObj);
        """
        web_driver_type = self.get_web_driver_type()
        try:
            try:
                if web_driver_type in [WebDriverType.CHROME, WebDriverType.FIREFOX]:
                    self.get_web_driver().execute_script(script, self)
                else:
                    ActionChains(self.get_web_driver()._selenium_web_driver()).move_to_element(
                        self._selenium_element()).perform()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                if web_driver_type in [WebDriverType.CHROME, WebDriverType.FIREFOX]:
                    self.get_web_driver().execute_script(script, self)
                else:
                    ActionChains(self.get_web_driver()._selenium_web_driver()).move_to_element(
                        self._selenium_element()).perform()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def click_and_hold(self):
        try:
            try:
                ActionChains(self.get_web_driver()._selenium_web_driver()).click_and_hold(
                    self._selenium_element()).perform()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                ActionChains(self.get_web_driver()._selenium_web_driver()).click_and_hold(
                    self._selenium_element()).perform()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def release_mouse_here(self):
        try:
            try:
                ActionChains(self.get_web_driver()._selenium_web_driver()).move_to_element(
                    self._selenium_element()).release().perform()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                ActionChains(self.get_web_driver()._selenium_web_driver()).move_to_element(
                    self._selenium_element()).release().perform()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._BROWSER)
    def release_mouse_here_with_offset(self, x_offset, y_offset):
        """
            Release mouse here with offset.
            The origin is at the top-left corner of web driver and offsets are relative to the top-left corner of the element.
        :param x_offset: X offset to release mouse.
        :param y_offset: Y offset to release mouse.
        """
        try:
            try:
                ActionChains(self.get_web_driver()._selenium_web_driver()).move_to_element_with_offset(
                    self._selenium_element(), x_offset, y_offset).release().perform()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                ActionChains(self.get_web_driver()._selenium_web_driver()).move_to_element_with_offset(
                    self._selenium_element(), x_offset, y_offset).release().perform()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def drag_and_drop_to(self, target_element):
        web_driver_type = self.get_web_driver_type()
        if web_driver_type in WebDriverType._MOBILE:
            try:
                try:
                    self.get_web_driver()._selenium_web_driver().drag_and_drop(self._selenium_element(),
                                                                               target_element._selenium_element())
                except (NoSuchElementException, StaleElementReferenceException):
                    self.wait_for().visible()
                    target_element.wait_for().visible()
                    self.get_web_driver()._selenium_web_driver().drag_and_drop(self._selenium_element(),
                                                                               target_element._selenium_element())
            except WebDriverException as wde:
                raise EasyiumException("%s\n%s" % (wde.msg, self))
        else:
            self.click_and_hold()
            target_element.release_mouse_here()

    @SupportedBy(WebDriverType._BROWSER)
    def drag_and_drop_to_with_offset(self, target_element, x_offset, y_offset):
        """
            Drag and drop to target element with offset.
            The origin is at the top-left corner of web driver and offsets are relative to the top-left corner of the element.
        :param target_element: the target element to drop.
        :param x_offset: X offset to drop
        :param y_offset: Y offset to drop.
        """
        self.click_and_hold()
        target_element.release_mouse_here_with_offset(x_offset, y_offset)

    @SupportedBy(WebDriverType._MOBILE)
    def tap(self):
        from appium.webdriver.common.touch_action import TouchAction

        touch_action = TouchAction(self.get_web_driver()._selenium_web_driver())
        try:
            try:
                touch_action.tap(self._selenium_element(), None, None, 1).perform()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                touch_action.tap(self._selenium_element(), None, None, 1).perform()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._MOBILE)
    def press(self):
        from appium.webdriver.common.touch_action import TouchAction

        touch_action = TouchAction(self.get_web_driver()._selenium_web_driver())
        try:
            try:
                touch_action.press(self._selenium_element(), None, None).release().perform()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                touch_action.press(self._selenium_element(), None, None).release().perform()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._MOBILE)
    def long_press(self, duration=1000):
        from appium.webdriver.common.touch_action import TouchAction

        touch_action = TouchAction(self.get_web_driver()._selenium_web_driver())
        try:
            try:
                touch_action.long_press(self._selenium_element(), None, None, duration).release().perform()
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                touch_action.long_press(self._selenium_element(), None, None, duration).release().perform()
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType.ANDROID)
    def set_text(self, text):
        try:
            try:
                self._selenium_element().set_text(text)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                self._selenium_element().set_text(text)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._MOBILE)
    def get_location_in_view(self):
        try:
            try:
                location = self._selenium_element().location_in_view
                return location["x"], location["y"]
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                location = self._selenium_element().location_in_view
                return location["x"], location["y"]
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._MOBILE)
    def set_value(self, value):
        try:
            try:
                self._selenium_element().set_value(value)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                self._selenium_element().set_value(value)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._MOBILE)
    def scroll_to(self, target_element):
        try:
            try:
                self.get_web_driver()._selenium_web_driver().scroll(self._selenium_element(),
                                                                    target_element._selenium_element())
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().exists()
                target_element.wait_for().exists()
                self.get_web_driver()._selenium_web_driver().scroll(self._selenium_element(),
                                                                    target_element._selenium_element())
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._MOBILE)
    def pinch(self, percent=200, steps=50):
        try:
            try:
                self.get_web_driver()._selenium_web_driver().pinch(self._selenium_element(), percent, steps)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                self.get_web_driver()._selenium_web_driver().pinch(self._selenium_element(), percent, steps)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    @SupportedBy(WebDriverType._MOBILE)
    def zoom(self, percent=200, steps=50):
        try:
            try:
                self.get_web_driver()._selenium_web_driver().zoom(self._selenium_element(), percent, steps)
            except (NoSuchElementException, StaleElementReferenceException):
                self.wait_for().visible()
                self.get_web_driver()._selenium_web_driver().zoom(self._selenium_element(), percent, steps)
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def is_displayed(self):
        try:
            try:
                return self._selenium_element().is_displayed()
            except StaleElementReferenceException:
                self._refresh()
                return self._selenium_element().is_displayed()
        except NoSuchElementException:
            return False
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))

    def exists(self):
        try:
            try:
                self._selenium_element().is_displayed()
                return True
            except StaleElementReferenceException:
                self._refresh()
                return True
        except NoSuchElementException:
            return False
        except WebDriverException as wde:
            raise EasyiumException("%s\n%s" % (wde.msg, self))



import time

from selenium.common.exceptions import NoAlertPresentException

from .exceptions import TimeoutException, NoSuchElementException
from .config import DEFAULT, default_config

__author__ = 'karl.gong'


class Waiter:
    def __init__(self, interval=DEFAULT, timeout=DEFAULT):
        """
            Create a Waiter instance.

        :param interval: the wait interval (in milliseconds), default value is from default_config.waiter_wait_interval
        :param timeout: the wait timeout (in milliseconds), default value is from default_config.waiter_wait_timeout
        """
        self.__interval = default_config.waiter_wait_interval if interval == DEFAULT else interval
        self.__timeout = default_config.waiter_wait_timeout if timeout == DEFAULT else timeout

    def wait_for(self, condition_function, *function_args, **function_kwargs):
        """
            Wait for the condition.

        :param condition_function: the condition function
        :param function_args: the args for condition_function
        :param function_kwargs: the kwargs for condition_function
        """
        start_time = time.time() * 1000.0

        if condition_function(*function_args, **function_kwargs):
            return

        while (time.time() * 1000.0 - start_time) <= self.__timeout:
            if condition_function(*function_args, **function_kwargs):
                return
            else:
                time.sleep(self.__interval / 1000.0)

        raise TimeoutException("Timed out waiting for <%s>." % condition_function.__name__)


class ElementWaitFor:
    def __init__(self, element, interval, timeout):
        self.__element = element
        self.__desired_occurrence = True
        self.__waiter = Waiter(interval, timeout)

    def __wait_for(self, element_condition):
        def is_element_condition_occurred():
            return element_condition.occurred() == self.__desired_occurrence

        try:
            self.__waiter.wait_for(is_element_condition_occurred)
        except TimeoutException:
            raise TimeoutException(
                "Timed out waiting for <%s> to be <%s>." % (element_condition, self.__desired_occurrence))

    def not_(self):
        self.__desired_occurrence = not self.__desired_occurrence
        return self

    def exists(self):
        self.__wait_for(ElementExistence(self.__element))

    def visible(self):
        self.__wait_for(ElementVisible(self.__element))

    def attribute_contains_one(self, attribute, *values):
        self.__wait_for(ElementAttributeContainsOne(self.__element, attribute, *values))

    def attribute_contains_all(self, attribute, *values):
        self.__wait_for(ElementAttributeContainsAll(self.__element, attribute, *values))


class ElementExistence:
    def __init__(self, element):
        self.__element = element

    def occurred(self):
        return self.__element.exists()

    def __str__(self):
        return "ElementExistence [\n%s\n]" % self.__element


class ElementVisible:
    def __init__(self, element):
        self.__element = element

    def occurred(self):
        return self.__element.is_displayed()

    def __str__(self):
        return "ElementVisible [\n%s\n]" % self.__element


class ElementAttributeContainsOne:
    def __init__(self, element, attribute, *values):
        self.__element = element
        self.__attribute = attribute
        self.__values = values

    def occurred(self):
        attribute_value = self.__element.get_attribute(self.__attribute)
        for value in self.__values:
            if attribute_value.find(value) != -1:
                return True
        return False

    def __str__(self):
        return "ElementAttributeContainsOne [element: \n%s\n][attribute: %s][values: %s]" % (
            self.__element, self.__attribute, self.__values)


class ElementAttributeContainsAll:
    def __init__(self, element, attribute, *values):
        self.__element = element
        self.__attribute = attribute
        self.__values = values

    def occurred(self):
        attribute_value = self.__element.get_attribute(self.__attribute)
        for value in self.__values:
            if attribute_value.find(value) == -1:
                return False
        return True

    def __str__(self):
        return "ElementAttributeContainsAll [element: \n%s\n][attribute: %s][values: %s]" % (
            self.__element, self.__attribute, self.__values)


class WebDriverWaitFor:
    def __init__(self, web_driver, interval, timeout):
        self.__web_driver = web_driver
        self.__desired_occurrence = True
        self.__waiter = Waiter(interval, timeout)

    def __wait_for(self, web_driver_condition):
        def is_web_driver_condition_occurred():
            return web_driver_condition.occurred() == self.__desired_occurrence

        try:
            self.__waiter.wait_for(is_web_driver_condition_occurred)
        except TimeoutException:
            raise TimeoutException(
                "Timed out waiting for <%s> to be <%s>." % (web_driver_condition, self.__desired_occurrence))

    def not_(self):
        self.__desired_occurrence = not self.__desired_occurrence
        return self

    def alert_present(self):
        self.__wait_for(AlertPresent(self.__web_driver))

    def text_present(self, text):
        self.__wait_for(TextPresent(self.__web_driver, text))


class AlertPresent:
    def __init__(self, web_driver):
        self.__web_driver = web_driver

    def occurred(self):
        try:
            alert_text = self.__web_driver.get_alert().text
            return True
        except NoAlertPresentException:
            return False

    def __str__(self):
        return "AlertPresent [\n%s\n]" % self.__web_driver


class TextPresent:
    def __init__(self, web_driver, text):
        self.__web_driver = web_driver
        self.__text = text

    def occurred(self):
        try:
            self.__web_driver.find_element("xpath=//*[contains(text(), '%s')]" % self.__text)
            return True
        except NoSuchElementException:
            return False

    def __str__(self):
        return "TextPresent [webdriver: \n%s\n][text: %s]" % (self.__web_driver, self.__text)
